@@ -2,6 +2,7 @@
 from __future__ import absolute_import, unicode_literals
 
 # Standard Library
+import json
 import re
 import sys
 from builtins import bytes  # noqa
@@ -25,9 +26,6 @@ class AMPNode:
         pass
 
     class Translator:
-        class TranslationError(Exception):
-            pass
-
         @classmethod
         def parse_sizes(cls, value):
             # Normalize whitespace
@@ -43,13 +41,15 @@ class AMPNode:
 
             Media = namedtuple('Media', 'query value')  # noqa
             for size_value in size_values:
-                size_parts = re.split(r'\s+', size_value)
-                if len(size_parts) <= 1:
-                    raise cls.TranslationError('Invalid sizes definition')
+                size_parts = re.split(r'\)\s+', size_value)
+                if len(size_parts) != 2:
+                    raise ValueError('Invalid sizes definition')
 
-                size_value = size_parts.pop()
-                media = Media(query=' '.join(size_parts), value=size_value)
+                query = '{})'.format(size_parts[0])
+                query = query.replace(r'\s+', '')
+                value = size_parts[1]
 
+                media = Media(query=query, value=value)
                 sizes.other.append(media)
 
             return sizes
@@ -187,6 +187,8 @@ class AMPNode:
         self.sizer = None
         self.maybe_img_attrs = None
 
+        self.should_strip_translated_attrs = True
+
         for attr in attrs:
             if attr[0] == 'id':
                 self.id = attr[1]
@@ -240,26 +242,44 @@ class AMPNode:
 
         translations = [k for k in self._other_attrs if k in self.TRANSLATIONS]
         if translations:
-            used_auto_id = False
-            if not self.id:
-                self.id = '{}{}'.format(self.ID_PREFIX, next_auto_id_num)
-                used_auto_id = True
+            potential_id = self.id or '{}{}'.format(self.ID_PREFIX, next_auto_id_num)
 
             css_parts = []
             for t in translations:
+                if t == 'sizes' and 'srcset' not in self._other_attrs:
+                    """According to the Mozilla docs, a sizes attribute without
+                    a valid srcset attribute should have no effect. Therefore,
+                    it should simply be stripped, without producing media
+                    queries.
+
+                    https://developer.mozilla.org/en-US/docs/Web/HTML/Element/img#attr-sizes  # noqa
+                    """
+
+                    continue
+
                 attribute_value = self._other_attrs[t]
                 Translator = self.TRANSLATIONS[t]  # noqa
+
                 try:
-                    css_part = Translator.make_translated_css(attribute_value, self.id)
-                except Translator.TranslationError:
-                    raise self.TransformationError('Error translating "{}" attribute to css')
-                css_parts.append(css_part)
+                    css_part = Translator.make_translated_css(attribute_value, potential_id)
+                except ValueError:
+                    raise self.TransformationError('Invalid value for `{}` attribute'.format(t))
+
+                if css_part:
+                    css_parts.append(css_part)
+
             css = ''.join(css_parts)
+            if css:
+                used_auto_id = False
+                if not self.id:
+                    used_auto_id = True
+                    self.id = potential_id
 
-            translation = css, used_auto_id
+                translation = css, used_auto_id
 
-            for t in translations:
-                del self._other_attrs[t]
+            if self.should_strip_translated_attrs:
+                for t in translations:
+                    del self._other_attrs[t]
 
         layout_value = self._other_attrs.get('layout')
 
@@ -314,10 +334,10 @@ class AMPNode:
             if not all(isinstance(length, self.CSSLength) for length in [width, height]):
                 raise self.TransformationError('Length and width required for fixed layout')
 
-            self._style = 'width:{wn}{wu};height:{hn}{hu}{existing_style};'.format(
-                wn=width.numeral,
+            self._style = 'width:{wn}{wu};height:{hn}{hu};{existing_style}'.format(
+                wn=str(width.numeral).rstrip('0').rstrip('.'),
                 wu=width.unit.value,
-                hn=height.numeral,
+                hn=str(height.numeral).rstrip('0').rstrip('.'),
                 hu=height.unit.value,
                 existing_style=self._style)
 
@@ -325,21 +345,21 @@ class AMPNode:
             if not isinstance(height, self.CSSLength):
                 raise self.TransformationError('Length and width required for fixed layout')
 
-            self._style = 'height:{numeral}{unit}{existing_style};'.format(
-                numeral=height.numeral,
+            self._style = 'height:{numeral}{unit};{existing_style}'.format(
+                numeral=str(height.numeral).rstrip('0').rstrip('.'),
                 unit=height.unit.value,
                 existing_style=self._style)
 
         elif layout == self.Layout.FLEX_ITEM:
             if isinstance(height, self.CSSLength):
-                self._style = 'height:{numeral}{unit}{existing_style};'.format(
-                    numeral=height.numeral,
+                self._style = 'height:{numeral}{unit};{existing_style}'.format(
+                    numeral=str(height.numeral).rstrip('0').rstrip('.'),
                     unit=height.unit.value,
                     existing_style=self._style)
 
             if isinstance(width, self.CSSLength):
-                self._style = 'width:{numeral}{unit}{existing_style};'.format(
-                    numeral=width.numeral,
+                self._style = 'width:{numeral}{unit};{existing_style}'.format(
+                    numeral=str(width.numeral).rstrip('0').rstrip('.'),
                     unit=width.unit.value,
                     existing_style=self._style)
 
@@ -370,7 +390,6 @@ class AMPNode:
             self.maybe_img_attrs = img_attrs
 
         # Create sizer if necessary
-
         if all(isinstance(length, self.CSSLength) for length in [width, height]):
             if all([width.numeral != 0, width.unit == height.unit]):
                 Sizer = namedtuple('Sizer', 'attrs maybe_img_attrs')
@@ -397,6 +416,10 @@ class AMPNode:
         return translation
 
     def get_attrs(self):
+        """Return an list of attribute tuples that represents current state.
+
+        This returns different values before and after a call to `transform()`.
+        """
         attrs = list(self._other_attrs.items())
 
         if self.id:
@@ -409,7 +432,7 @@ class AMPNode:
             attrs.insert(0, ('style', self._style))
 
         if self._is_hidden:
-            attrs.append(('hidden', None))
+            attrs.append(('hidden', 'hidden'))
 
         return attrs
 
@@ -417,18 +440,17 @@ class AMPNode:
 class AMPRenderer(HTMLParser, object):
     """A parser to ingest AMP HTML and perform various transformations."""
 
+    # These are added in where found, and removed during cleanup.
     TRANSLATED_STYLES_PLACEHOLDER = '/* style-amp-custom-translated */'
     BOILERPLATE_PLACEHOLDER = '/* style-amp-boilerplate */'
     NOSCRIPT_BOILERPLATE_PLACEHOLDER = '/* style-amp-boilerplate-noscript */'
 
     RENDER_DELAYING_EXTENSIONS = [
         'amp-dynamic-css-classes',
-        'amp-experiment',
+        # 'amp-experiment',  # we handle this by looking for the tag instead
         'amp-story',
     ]
 
-    _next_auto_id_num = 1
-
     def __init__(self, runtime_styles, runtime_version, *args, **kwargs):
         """Initialize AMPRenderer with runtime styles & version.
 
@@ -446,6 +468,10 @@ class AMPRenderer(HTMLParser, object):
         self.runtime_styles = runtime_styles
         self.runtime_version = runtime_version
 
+        self._is_test_mode = False
+        if not self.runtime_styles or not self.runtime_version:
+            self._is_test_mode = True
+
         self.should_trim_attrs = False
         self.should_strip_comments = False
 
@@ -463,15 +489,58 @@ class AMPRenderer(HTMLParser, object):
         self._noscript_boilerplate = ''
         self._is_in_noscript = False
 
+        self._is_expecting_experiment_script = False
+        self._is_expecting_experiment_data = False
+        self._is_expecting_experiment_end = False
+        self._current_experiment_data = ''
+
         self._is_render_paused = False
+        self._is_render_cancelled = False
 
         self._result = ''
+        self._found_custom_element = False
+        self._next_auto_id_num = 0
+        self._translated_styles = ''
+
+        try:
+            del self.no_boilerplate
+        except AttributeError:
+            pass
+
+    def _apply_experiment_data(self):
+        self._result = '{}{}'.format(self._result, self._current_experiment_data)
+
+        experiment_data = self._current_experiment_data
+
+        self._current_experiment_data = ''
+        self._is_expecting_experiment_data = False
+
+        return experiment_data
 
     def handle_decl(self, decl):
         self._result = '{}<!{}>'.format(self._result, decl.lower())
 
     def handle_starttag(self, tag, attrs):
         tag = tag.lower()
+
+        if self._is_expecting_experiment_data:
+            # Expecting some JSON, found a start tag == NO EXPERIMENT
+            self._apply_experiment_data()
+
+        if self._is_expecting_experiment_script:
+            self._is_expecting_experiment_script = False
+
+            if tag == 'script':
+                for attr in attrs:
+                    if attr[0] == 'type' and attr[1] == 'application/json':
+                        """Expecting script & found it; Next, we expect some
+                        JSON data"""
+                        self._is_expecting_experiment_data = True
+
+        if self._is_expecting_experiment_end:
+            # Expecting </amp-experiment>, found a start tag == NO EXPERIMENT
+            self._is_expecting_experiment_end = False
+
         if tag == 'noscript':
             self._is_in_noscript = True
 
@@ -479,6 +548,7 @@ class AMPRenderer(HTMLParser, object):
             if 'amp-boilerplate' in (attr[0] for attr in attrs):
                 self._is_in_boilerplate = True
 
+                # Add appropriate boilerplate placeholder
                 if self._is_in_noscript and self.NOSCRIPT_BOILERPLATE_PLACEHOLDER not in self._result:
                     self._result = '{}{}'.format(self._result, self.NOSCRIPT_BOILERPLATE_PLACEHOLDER)
                 elif self.BOILERPLATE_PLACEHOLDER not in self._result:
@@ -486,25 +556,48 @@ class AMPRenderer(HTMLParser, object):
 
                 return
 
+        safe_attrs = attrs
+
+        if tag == 'html':
+            if 'i-amphtml-layout' in (attr[0] for attr in attrs):
+                # A simple check to see if it’s transformed already
+                self._is_render_cancelled = True
+            else:
+                html_attrs = [('i-amphtml-layout', None), ('i-amphtml-no-boilerplate', None)]
+                if not self._is_test_mode:
+                    html_attrs.append(('transformed', 'self;v=1'))
+
+                safe_attrs.extend(html_attrs)
+
         if tag in ['template', 'script']:
+            # Don't render amp-* elements inside of `template` or `script`
             self._is_render_paused = True
 
-        if tag == 'script':
-            for attr in attrs:
-                if attr[0] == 'custom-element' and attr[1] in self.RENDER_DELAYING_EXTENSIONS:
-                    self._should_remove_boilerplate = False
+            if tag == 'script':
+                for attr in attrs:
+                    if attr[0] == 'custom-element' and attr[1] in self.RENDER_DELAYING_EXTENSIONS:
+                        """Don’t remove boilerplate if one of the render-
+                        delaying extensions is included"""
+                        self._should_remove_boilerplate = False
 
         sizer = None
         maybe_img_attrs = None
         if tag == 'amp-audio':
+            # Don’t remove boilerplate if `amp-audio` is included
             self._should_remove_boilerplate = False
-            safe_attrs = attrs
 
-        elif not self._is_render_paused and tag.startswith('amp-'):
+        elif not self._is_render_cancelled and not self._is_render_paused and tag.startswith('amp-'):
             if tag == 'amp-experiment':
-                self._should_remove_boilerplate = False
+                """Start the finite automata to see if we need to keep the
+                boilerplate"""
+                self._is_expecting_experiment_script = True
 
             amp_element = AMPNode(tag, attrs)
+            if not self._is_test_mode:
+                """AMP Optimizer strips these out, and I believe this is
+                incorrect, because the runtime handles them smartly when
+                left alone."""
+                amp_element.should_strip_translated_attrs = False
 
             try:
                 transformation = amp_element.transform(self._next_auto_id_num)
@@ -513,25 +606,17 @@ class AMPRenderer(HTMLParser, object):
             else:
                 if transformation:
                     css, used_auto_id = transformation
-                    placeholder = self.TRANSLATED_STYLES_PLACEHOLDER
-                    self._result = self._result.replace(placeholder, '{}{}'.format(css, placeholder))
+                    self._translated_styles = '{}{}'.format(self._translated_styles, css)
 
                     if used_auto_id:
+                        # We had to generate an ID for this element
                         self._next_auto_id_num += 1
 
             safe_attrs = amp_element.get_attrs()
             sizer = amp_element.sizer
             maybe_img_attrs = amp_element.maybe_img_attrs
 
-        else:
-            safe_attrs = attrs
-
-        if tag == 'html':
-            for attr in [('i-amphtml-layout', None),
-                         ('i-amphtml-no-boilerplate', None),
-                         ('transformed', 'self;v=1')]:
-                safe_attrs.append(attr)
-
+        # Turn attribute data in to strings
         attr_strings = []
         for attr in safe_attrs:
             if attr[1] is not None:
@@ -543,10 +628,15 @@ class AMPRenderer(HTMLParser, object):
             else:
                 attr_strings.append(' {}'.format(attr[0].lower()))
 
+        if self._is_test_mode:
+            # Sort alphabetically for diffing
+            attr_strings.sort()
+
         attr_string = ''.join(attr_strings)
 
         self._result = '{}<{}{}>'.format(self._result, tag, attr_string)
 
+        # Add sizer if necessary
         if sizer:
             sizer_attr_strings = []
             for attr in sizer.attrs:
@@ -575,6 +665,7 @@ class AMPRenderer(HTMLParser, object):
 
             self._result = '{}</i-amphtml-sizer>'.format(self._result)
 
+        # Add img if necessary
         if maybe_img_attrs:
             img_attr_strings = []
             for attr in maybe_img_attrs:
@@ -588,18 +679,56 @@ class AMPRenderer(HTMLParser, object):
 
             self._result = '{}<img{}>'.format(self._result, img_attr_string)
 
-        if tag == 'head':
-            style = '<style amp-runtime i-amphtml-version="{}">{}</style>'.format(
-                self.runtime_version,
-                self.runtime_styles)
+        # Add runtime styles if necessary
+        if tag == 'head' and not self._is_render_cancelled:
+            if self._is_test_mode:
+                """AMP Optimizer uses a stub like this, and then replaces it
+                later. Use a stub so we can test against their expected
+                output."""
+                style = '<style amp-runtime></style>'
+            else:
+                style = '<style amp-runtime i-amphtml-version="{}">{}</style>'.format(
+                    self.runtime_version,
+                    self.runtime_styles)
             self._result = '{}{}'.format(self._result, style)
 
         if tag == 'style':
+            """Insert a placeholder into <style amp-custom> so we can add in
+            the transformed styles later."""
             if 'amp-custom' in (attr[0] for attr in attrs) and self.TRANSLATED_STYLES_PLACEHOLDER not in self._result:
+                self._found_custom_element = True
                 self._result = '{}{}'.format(self._result, self.TRANSLATED_STYLES_PLACEHOLDER)
 
     def handle_endtag(self, tag):
         tag = tag.lower()
+
+        if self._is_expecting_experiment_data:
+            # Finish ingesting JSON data
+            data = self._apply_experiment_data()
+
+            if tag == 'script':
+                try:
+                    # If valid JSON...
+                    json_data = json.loads(data)
+                except ValueError:
+                    pass
+                else:
+                    # If data wasn’t empty...
+                    if json_data:
+                        self._is_expecting_experiment_end = True
+
+        elif self._is_expecting_experiment_script:
+            self._is_expecting_experiment_script = False
+
+        elif self._is_expecting_experiment_end:
+            """If successful experiment and only one child of node, then
+            there is an experiment active and the boilerplate can't be
+            removed."""
+            self._is_expecting_experiment_end = False
+
+            if tag == 'amp-experiment':
+                self._should_remove_boilerplate = False
+
         if tag == 'noscript':
             self._is_in_noscript = False
 
@@ -607,6 +736,15 @@ class AMPRenderer(HTMLParser, object):
             self._is_in_boilerplate = False
             return
 
+        if tag == 'head' and not self._found_custom_element and self.TRANSLATED_STYLES_PLACEHOLDER not in self._result:
+            """If there was no custom element found in the head, add the
+            placeholder at the end in case we have custom styles to add later.
+
+            self._found_custom_element will remain False, and we’ll inspect
+            that later to decide whether the <script> element itself needs to
+            be added."""
+            self._result = '{}{}'.format(self._result, self.TRANSLATED_STYLES_PLACEHOLDER)
+
         if tag in ['template', 'script']:
             self._is_render_paused = False
 
@@ -621,6 +759,10 @@ class AMPRenderer(HTMLParser, object):
             self._boilerplate = '{}{}'.format(self._boilerplate, data)
             return
 
+        if self._is_expecting_experiment_data:
+            self._current_experiment_data = '{}{}'.format(self._current_experiment_data, data)
+            return
+
         self._result = '{}{}'.format(self._result, data)
 
     def handle_data(self, data):
@@ -641,16 +783,28 @@ class AMPRenderer(HTMLParser, object):
         self.feed(data)
         self.close()
 
-        self._result = self._result.replace(self.TRANSLATED_STYLES_PLACEHOLDER, '')
+        style_string = self._translated_styles
+        if style_string and not self._found_custom_element:
+            # Insert the amp-custom tag if necessary
+            style_string = '<style amp-custom>{}</style>'.format(style_string)
+        self._result = self._result.replace(self.TRANSLATED_STYLES_PLACEHOLDER, style_string)
 
         boilerplate = ''
         noscript_boilerplate = ''
-        if not self._should_remove_boilerplate:
-            boilerplate = self._boilerplate
-            noscript_boilerplate = self._noscript_boilerplate
+
+        self.no_boilerplate = True
+        if self._is_render_cancelled or not self._should_remove_boilerplate:
+            self.no_boilerplate = False
+
+            # Restore the boilerplate
+            boilerplate = '<style amp-boilerplate>{}</style>'.format(self._boilerplate)
+            noscript_boilerplate = '<style amp-boilerplate>{}</style>'.format(self._noscript_boilerplate)
             self._result = self._result.replace(' i-amphtml-no-boilerplate', '')
 
         self._result = self._result.replace(self.BOILERPLATE_PLACEHOLDER, boilerplate)
         self._result = self._result.replace(self.NOSCRIPT_BOILERPLATE_PLACEHOLDER, noscript_boilerplate)
 
+        # Remove empty noscript tags; This happens when removing boilerplate
+        self._result = self._result.replace('<noscript></noscript>', '')
+
         return self._result
